from __future__ import absolute_import

import base64
import contextlib
import time
import typing as tp

from PIL import Image
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.switch_to import SwitchTo
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait

from applitools.core import logger
from applitools.core.errors import EyesError
from applitools.core.geometry import Point, Region
from applitools.utils import cached_property, image_utils, general_utils
from . import eyes_selenium_utils, StitchMode
from .positioning import ElementPositionProvider, build_position_provider_for, ScrollPositionProvider
from .webelement import EyesWebElement
from .frames import Frame, FrameChain

if tp.TYPE_CHECKING:
    from applitools.core.scaling import ScaleProvider
    from applitools.utils.custom_types import Num, ViewPort, FrameReference, AnyWebDriver, AnyWebElement
    from .eyes import Eyes


class FrameResolver(object):
    def __init__(self, frame_ref, driver):
        # Find the frame's location and add it to the current driver offset
        if isinstance(frame_ref, str):
            frame_eyes_webelement = driver.find_element_by_name(frame_ref)
        elif isinstance(frame_ref, int):
            frame_elements_list = driver.find_elements_by_css_selector('frame, iframe')
            frame_eyes_webelement = frame_elements_list[frame_ref]
        elif isinstance(frame_ref, EyesWebElement):
            frame_eyes_webelement = frame_ref
        else:
            # It must be a WebElement
            frame_eyes_webelement = EyesWebElement(frame_ref, driver)

        self.eyes_webelement = frame_eyes_webelement
        self.webelement = frame_eyes_webelement.element


class _EyesSwitchTo(object):
    """
    Wraps a selenium "SwitchTo" object, so we can keep track of switching between frames.
    It has name EyesTargetLocator in other SDK's
    """
    # TODO: Make more similar to EyesTargetLocator
    _READONLY_PROPERTIES = ['alert', 'active_element']
    PARENT_FRAME = 1

    def __init__(self, driver, switch_to):
        # type: (EyesWebDriver, SwitchTo) -> None
        """
        Ctor.

        :param driver: EyesWebDriver instance.
        :param switch_to: Selenium switchTo object.
        """
        self._switch_to = switch_to
        self._driver = driver
        self._scroll_position = ScrollPositionProvider(driver)
        general_utils.create_proxy_interface(self, switch_to, self._READONLY_PROPERTIES)

    @contextlib.contextmanager
    def frame_and_back(self, frame_reference):
        # type: (FrameReference) -> tp.Generator
        self.frame(frame_reference)
        yield
        self.parent_frame()

    def frame(self, frame_reference):
        # type: (FrameReference) -> None
        """
        Switch to a given frame.

        :param frame_reference: The reference to the frame.
        """
        frame = FrameResolver(frame_reference, self._driver)
        self.will_switch_to_frame(frame.eyes_webelement)
        self._switch_to.frame(frame.webelement)

    def frames(self, frame_chain):
        # type: (FrameChain) -> None
        """
        Switches to the frames one after the other.

        :param frame_chain: A list of frames.
        """
        self._switch_to.default_content()
        for frame in frame_chain:
            self.frame(frame.reference)

    def default_content(self):
        # type: () -> None
        """
        Switch to default content.
        """
        self._driver.frame_chain.clear()
        self._switch_to.default_content()

    def parent_frame(self):
        """
        Switch to parent frame.
        """
        frames = self._driver.frame_chain
        if frames:
            frames.pop()

            try:
                self._switch_to.parent_frame()
            except WebDriverException as e:
                self._switch_to.default_content()
                for frame in frames:
                    self._switch_to.frame(frame.reference)

    def window(self, window_name):
        # type: (tp.Text) -> None
        """
        Switch to window.
        """
        self._driver.frame_chain.clear()
        self._switch_to.window(window_name)

    def will_switch_to_frame(self, target_frame):
        # type: (EyesWebElement) -> None
        """
        Will be called before switching into a frame.

        :param target_frame: The element about to be switched to.
        """
        assert target_frame is not None
        pl = target_frame.location

        size_and_borders = target_frame.size_and_borders
        borders = size_and_borders.borders
        frame_inner_size = size_and_borders.size

        content_location = Point(pl['x'] + borders['left'], pl['y'] + borders['top'])
        original_location = self._scroll_position.get_current_position()

        self._driver.scroll_to(content_location)
        frame = Frame(target_frame, content_location, target_frame.size, frame_inner_size,
                      parent_scroll_position=original_location)
        self._driver.frame_chain.push(frame)


class EyesWebDriver(object):
    """
    A wrapper for selenium web driver which creates wrapped elements, and notifies us about
    events / actions.
    """
    # Properties require special handling since even testing if they're callable "activates"
    # them, which makes copying them automatically a problem.
    _READONLY_PROPERTIES = ['application_cache', 'current_url', 'current_window_handle',
                            'desired_capabilities', 'log_types', 'name', 'page_source', 'title',
                            'window_handles', 'switch_to', 'mobile', 'current_context', 'context',
                            'current_activity', 'network_connection', 'available_ime_engines',
                            'active_ime_engine', 'device_time', 'w3c', 'contexts', 'current_package',
                            # Appium specific
                            'battery_info']
    _SETTABLE_PROPERTIES = ['orientation', 'file_detector']

    # This should pretty much cover all scroll bars (and some fixed position footer elements :) ).
    _MAX_SCROLL_BAR_SIZE = 50

    _MIN_SCREENSHOT_PART_HEIGHT = 10

    def __init__(self, driver, eyes, stitch_mode=StitchMode.Scroll):
        # type: (WebDriver, Eyes, tp.Text) -> None
        """
        Ctor.

        :param driver: remote WebDriver instance.
        :param eyes: A Eyes sdk instance.
        :param stitch_mode: How to stitch a page (default is with scrolling).
        """
        self.driver = driver
        self._eyes = eyes
        self._origin_position_provider = build_position_provider_for(StitchMode.Scroll, driver)
        self._position_provider = build_position_provider_for(stitch_mode, driver)
        # tp.List of frames the user switched to, and the current offset, so we can properly
        # calculate elements' coordinates
        self._frame_chain = FrameChain()
        self._default_content_viewport_size = None  # type: tp.Optional[ViewPort]

        self.driver_takes_screenshot = driver.capabilities.get('takesScreenshot', False)

        # Creating the rest of the driver interface by simply forwarding it to the underlying
        # driver.
        general_utils.create_proxy_interface(self, driver,
                                             self._READONLY_PROPERTIES + self._SETTABLE_PROPERTIES)

        for attr in self._READONLY_PROPERTIES:
            if not hasattr(self.__class__, attr):
                setattr(self.__class__, attr, general_utils.create_proxy_property(attr, 'driver'))
        for attr in self._SETTABLE_PROPERTIES:
            if not hasattr(self.__class__, attr):
                setattr(self.__class__, attr, general_utils.create_proxy_property(attr, 'driver', True))

    def get_display_rotation(self):
        # type: () -> int
        """
        Get the rotation of the screenshot.

        :return: The rotation of the screenshot we get from the webdriver in (degrees).
        """
        if self.platform_name == 'Android' and self.driver.orientation == "LANDSCAPE":
            return -90
        return 0

    def get_platform_name(self):
        return self.platform_name

    def get_platform_version(self):
        return self.platform_version

    @cached_property
    def platform_name(self):
        # type: () -> tp.Optional[tp.Text]
        return self.driver.desired_capabilities.get('platformName', None)

    @cached_property
    def platform_version(self):
        # type: () -> tp.Optional[tp.Text]
        return self.driver.desired_capabilities.get('platformVersion', None)

    @cached_property
    def browser_version(self):
        # type: () -> tp.Optional[float]
        caps = self.driver.capabilities
        version = caps.get('browserVersion', caps.get('version', None))
        if version:
            # convert version that has few dots in to float number (e.g. Edge 1.23.45)
            if version.find('.') != -1:
                version = float(version[:version.index('.') + 2])
            else:
                version = float(version)
        return version

    @cached_property
    def browser_name(self):
        # type: () -> tp.Optional[tp.Text]
        caps = self.driver.capabilities
        return caps.get('browserName', caps.get('browser', None))

    @cached_property
    def user_agent(self):
        try:
            user_agent = self.driver.execute_script("return navigator.userAgent")
            logger.info("user agent: {}".format(user_agent))
        except Exception as e:
            logger.info("Failed to obtain user-agent string")
            user_agent = None
        return user_agent

    def is_mobile_device(self):
        # type: () -> bool
        """
        Returns whether the platform running is a mobile device or not.

        :return: True if the platform running the test is a mobile platform. False otherwise.
        """
        return eyes_selenium_utils.is_mobile_device(self.driver)

    def get(self, url):
        # type: (tp.Text) -> tp.Optional[tp.Any]
        """
        Navigates the driver to the given url.

        :param url: The url to navigate to.
        :return: A driver that navigated to the given url.
        """
        # We're loading a new page, so the frame location resets
        self._frame_chain.clear()
        return self.driver.get(url)

    def find_element(self, by=By.ID, value=None):
        # type: (tp.Text, tp.Text) -> EyesWebElement
        """
        Returns a WebElement denoted by "By".

        :param by: By which option to search for (default is by ID).
        :param value: The value to search for.
        :return: A element denoted by "By".
        """
        # Get result from the original implementation of the underlying driver.
        result = self.driver.find_element(by, value)
        # Wrap the element.
        if result:
            result = EyesWebElement(result, self)
        return result

    def find_elements(self, by=By.ID, value=None):
        # type: (tp.Text, tp.Text) -> tp.List[EyesWebElement]
        """
        Returns a list of web elements denoted by "By".

        :param by: By which option to search for (default is by ID).
        :param value: The value to search for.
        :return: List of elements denoted by "By".
        """
        # Get result from the original implementation of the underlying driver.
        results = self.driver.find_elements(by, value)
        # Wrap all returned elements.
        if results:
            updated_results = []
            for element in results:
                updated_results.append(EyesWebElement(element, self))
            results = updated_results
        return results

    def find_element_by_id(self, id_):
        # type: (tp.Text) -> EyesWebElement
        """
        Finds an element by id.

        :params id_: The id of the element to be found.
        """
        return self.find_element(by=By.ID, value=id_)

    def find_elements_by_id(self, id_):
        # type: (tp.Text) -> tp.List[EyesWebElement]
        """
        Finds multiple elements by id.

        :param id_: The id of the elements to be found.
        """
        return self.find_elements(by=By.ID, value=id_)

    def find_element_by_xpath(self, xpath):
        # type: (tp.Text) -> EyesWebElement
        """
        Finds an element by xpath.

        :param xpath: The xpath locator of the element to find.
        """
        return self.find_element(by=By.XPATH, value=xpath)

    def find_elements_by_xpath(self, xpath):
        # type: (tp.Text) -> tp.List[EyesWebElement]
        """
        Finds multiple elements by xpath.

        :param xpath: The xpath locator of the elements to be found.
        """
        return self.find_elements(by=By.XPATH, value=xpath)

    def find_element_by_link_text(self, link_text):
        # type: (tp.Text) -> EyesWebElement
        """
        Finds an element by link text.

        :param link_text: The text of the element to be found.
        """
        return self.find_element(by=By.LINK_TEXT, value=link_text)

    def find_elements_by_link_text(self, text):
        # type: (tp.Text) -> tp.List[EyesWebElement]
        """
        Finds elements by link text.

        :param text: The text of the elements to be found.
        """
        return self.find_elements(by=By.LINK_TEXT, value=text)

    def find_element_by_partial_link_text(self, link_text):
        # type: (tp.Text) -> EyesWebElement
        """
        Finds an element by a partial match of its link text.

        :param link_text: The text of the element to partially match on.
        """
        return self.find_element(by=By.PARTIAL_LINK_TEXT, value=link_text)

    def find_elements_by_partial_link_text(self, link_text):
        # type: (tp.Text) -> tp.List[EyesWebElement]
        """
        Finds elements by a partial match of their link text.

        :param link_text: The text of the element to partial match on.
        """
        return self.find_elements(by=By.PARTIAL_LINK_TEXT, value=link_text)

    def find_element_by_name(self, name):
        # type: (tp.Text) -> EyesWebElement
        """
        Finds an element by name.

        :param name: The name of the element to find.
        """
        return self.find_element(by=By.NAME, value=name)

    def find_elements_by_name(self, name):
        # type: (tp.Text) -> tp.List[EyesWebElement]
        """
        Finds elements by name.

        :param name: The name of the elements to find.
        """
        return self.find_elements(by=By.NAME, value=name)

    def find_element_by_tag_name(self, name):
        # type: (tp.Text) -> EyesWebElement
        """
        Finds an element by tag name.

        :param name: The tag name of the element to find.
        """
        return self.find_element(by=By.TAG_NAME, value=name)

    def find_elements_by_tag_name(self, name):
        # type: (tp.Text) -> tp.List[EyesWebElement]
        """
        Finds elements by tag name.

        :param name: The tag name to use when finding elements.
        """
        return self.find_elements(by=By.TAG_NAME, value=name)

    def find_element_by_class_name(self, name):
        # type: (tp.Text) -> EyesWebElement
        """
        Finds an element by class name.

        :param name: The class name of the element to find.
        """
        return self.find_element(by=By.CLASS_NAME, value=name)

    def find_elements_by_class_name(self, name):
        # type: (tp.Text) -> tp.List[EyesWebElement]
        """
        Finds elements by class name.

        :param name: The class name of the elements to find.
        """
        return self.find_elements(by=By.CLASS_NAME, value=name)

    def find_element_by_css_selector(self, css_selector):
        # type: (tp.Text) -> EyesWebElement
        """
        Finds an element by css selector.

        :param css_selector: The css selector to use when finding elements.
        """
        return self.find_element(by=By.CSS_SELECTOR, value=css_selector)

    def find_elements_by_css_selector(self, css_selector):
        # type: (tp.Text) -> tp.List[EyesWebElement]
        """
        Finds elements by css selector.

        :param css_selector: The css selector to use when finding elements.
        """
        return self.find_elements(by=By.CSS_SELECTOR, value=css_selector)

    def get_screenshot_as_base64(self):
        # type: () -> tp.Text
        """
        Gets the screenshot of the current window as a base64 encoded string
           which is useful in embedded images in HTML.
        """
        screenshot64 = self.driver.get_screenshot_as_base64()
        display_rotation = self.get_display_rotation()
        if display_rotation != 0:
            logger.info('Rotation required.')
            # num_quadrants = int(-(display_rotation / 90))
            logger.debug('Done! Creating image object...')
            screenshot = image_utils.image_from_base64(screenshot64)

            # rotating
            if display_rotation == -90:
                screenshot64 = image_utils.get_base64(screenshot.rotate(90))
            logger.debug('Done! Rotating...')

        return screenshot64

    def get_screesnhot_as_base64_from_main_frame(self):
        # type: () -> tp.Text
        """
        Make screenshot from main frame
        """
        original_frame = self.frame_chain.clone()
        self.switch_to.default_content()
        screenshot64 = self.get_screenshot_as_base64()
        self.switch_to.frames(original_frame)
        return screenshot64

    def extract_full_page_width(self):
        # type: () -> int
        """
        Extracts the full page width.

        :return: The width of the full page.
        """
        # noinspection PyUnresolvedReferences
        default_scroll_width = int(round(self.driver.execute_script(
                "return document.documentElement.scrollWidth")))
        body_scroll_width = int(round(self.driver.execute_script("return document.body.scrollWidth")))
        return max(default_scroll_width, body_scroll_width)

    def extract_full_page_height(self):
        # type: () -> int
        """
        Extracts the full page height.

        :return: The height of the full page.
        IMPORTANT: Notice there's a major difference between scrollWidth and scrollHeight.
        While scrollWidth is the maximum between an element's width and its content width,
        scrollHeight might be smaller(!) than the clientHeight, which is why we take the
        maximum between them.
        """
        # noinspection PyUnresolvedReferences
        default_client_height = int(round(self.driver.execute_script(
                "return document.documentElement.clientHeight")))
        # noinspection PyUnresolvedReferences
        default_scroll_height = int(round(self.driver.execute_script(
                "return document.documentElement.scrollHeight")))
        # noinspection PyUnresolvedReferences
        body_client_height = int(round(self.driver.execute_script("return document.body.clientHeight")))
        # noinspection PyUnresolvedReferences
        body_scroll_height = int(round(self.driver.execute_script("return document.body.scrollHeight")))
        max_document_element_height = max(default_client_height, default_scroll_height)
        max_body_height = max(body_client_height, body_scroll_height)
        return max(max_document_element_height, max_body_height)

    def get_current_position(self):
        # type: () -> Point
        """
        Extracts the current scroll position from the browser.

        :return: The scroll position.
        """
        return self._origin_position_provider.get_current_position()

    def scroll_to(self, point):
        # type: (Point) -> None
        """
        Commands the browser to scroll to a given position.

        :param point: The point to scroll to.
        """
        self._origin_position_provider.set_position(point)

    def get_entire_page_size(self):
        # type: () -> tp.Dict[tp.Text, int]
        """
        Extracts the size of the current page from the browser using Javascript.

        :return: The page width and height.
        """
        return {'width':  self.extract_full_page_width(),
                'height': self.extract_full_page_height()}

    def set_overflow(self, overflow, stabilization_time=None):
        # type: (tp.Text, tp.Optional[int]) -> tp.Text
        """
        Sets the overflow of the current context's document element.

        :param overflow: The overflow value to set. If the given value is None, then overflow will be set to
                         undefined.
        :param stabilization_time: The time to wait for the page to stabilize after overflow is set. If the value is
                                    None, then no waiting will take place. (Milliseconds)
        :return: The previous overflow value.
        """
        logger.debug("Setting overflow: %s" % overflow)
        if overflow is None:
            script = "var origOverflow = document.documentElement.style.overflow; " \
                     "document.documentElement.style.overflow = undefined; " \
                     "return origOverflow;"
        else:
            script = "var origOverflow = document.documentElement.style.overflow; " \
                     "document.documentElement.style.overflow = \"{0}\"; " \
                     "return origOverflow;".format(overflow)
        # noinspection PyUnresolvedReferences
        original_overflow = self.driver.execute_script(script)
        logger.debug("Original overflow: %s" % original_overflow)
        if stabilization_time is not None:
            time.sleep(stabilization_time / 1000)
        return original_overflow

    def wait_for_page_load(self, timeout=3, throw_on_timeout=False):
        # type: (int, bool) -> None
        """
        Waits for the current document to be "loaded".

        :param timeout: The maximum time to wait, in seconds.
        :param throw_on_timeout: Whether to throw an exception when timeout is reached.
        """
        # noinspection PyBroadException
        try:
            WebDriverWait(self.driver, timeout) \
                .until(lambda driver: driver.execute_script('return document.readyState') == 'complete')
        except Exception:
            logger.debug('Page load timeout reached!')
            if throw_on_timeout:
                raise

    def hide_scrollbars(self):
        # type: () -> tp.Text
        """
        Hides the scrollbars of the current context's document element.

        :return: The previous value of the overflow property (could be None).
        """
        logger.debug('HideScrollbars() called. Waiting for page load...')
        self.wait_for_page_load()
        logger.debug('About to hide scrollbars')
        return self.set_overflow('hidden')

    @property
    def frame_chain(self):
        """
        Gets the frame chain.

        :return: A list of Frame instances which represents the path to the current frame.
            This can later be used as an argument to _EyesSwitchTo.frames().
        """
        return self._frame_chain

    def get_viewport_size(self):
        # type: () -> ViewPort
        """
        Returns:
            The viewport size of the current frame.
        """
        return eyes_selenium_utils.get_viewport_size(self)

    def get_default_content_viewport_size(self, force_query=False):
        # type: (bool) -> ViewPort
        """
        Gets the viewport size.

        :return: The viewport size of the most outer frame.
        """
        if self._default_content_viewport_size and not force_query:
            return self._default_content_viewport_size

        current_frames = self.frame_chain.clone()
        # If we're inside a frame, then we should first switch to the most outer frame.
        # Optimization
        if current_frames:
            self.switch_to.default_content()
        self._default_content_viewport_size = eyes_selenium_utils.get_viewport_size_or_display_size(self.driver)

        if current_frames:
            self.switch_to.frames(current_frames)

        return self._default_content_viewport_size

    def reset_origin(self):
        # type: () -> None
        """
        Reset the origin position to (0, 0).

        :raise EyesError: Couldn't scroll to position (0, 0).
        """
        self._origin_position_provider.push_state()
        self._origin_position_provider.set_position(Point(0, 0))
        current_scroll_position = self._origin_position_provider.get_current_position()
        if current_scroll_position.x != 0 or current_scroll_position.y != 0:
            self._origin_position_provider.pop_state()
            raise EyesError("Couldn't scroll to the top/left part of the screen!")

    def restore_origin(self):
        # type: () -> None
        """
        Restore the origin position.
        """
        self._origin_position_provider.pop_state()

    def save_position(self):
        """
        Saves the position in the _position_provider list.
        """
        self._position_provider.push_state()

    def restore_position(self):
        """
        Restore the position.
        """
        self._position_provider.pop_state()

    @staticmethod
    def _wait_before_screenshot(seconds):
        logger.debug("Waiting {} ms before taking screenshot..".format(int(seconds * 1000)))
        time.sleep(seconds)
        logger.debug("Finished waiting!")

    def get_full_page_screenshot(self, wait_before_screenshots, scale_provider):
        # type: (Num, ScaleProvider) -> Image.Image
        """
        Gets a full page screenshot.

        :param wait_before_screenshots: Seconds to wait before taking each screenshot.
        :return: The full page screenshot.
        """
        logger.info('getting full page screenshot..')

        # Saving the current frame reference and moving to the outermost frame.
        original_frame = self.frame_chain
        self.switch_to.default_content()

        self.reset_origin()

        entire_page_size = self.get_entire_page_size()

        # Starting with the screenshot at 0,0
        EyesWebDriver._wait_before_screenshot(wait_before_screenshots)
        part64 = self.get_screenshot_as_base64()
        screenshot = image_utils.image_from_bytes(base64.b64decode(part64))

        scale_provider.update_scale_ratio(screenshot.width)
        pixel_ratio = 1.0 / scale_provider.scale_ratio
        need_to_scale = True if pixel_ratio != 1.0 else False
        if need_to_scale:
            screenshot = image_utils.scale_image(screenshot, 1.0 / pixel_ratio)

        # IMPORTANT This is required! Since when calculating the screenshot parts for full size,
        # we use a screenshot size which is a bit smaller (see comment below).
        if (screenshot.width >= entire_page_size['width']) and \
                (screenshot.height >= entire_page_size['height']):
            self.restore_origin()
            self.switch_to.frames(original_frame)

            return screenshot

        #  We use a smaller size than the actual screenshot size in order to eliminate duplication
        #  of bottom scroll bars, as well as footer-like elements with fixed position.
        screenshot_part_size = {'width':  screenshot.width,
                                'height': max(screenshot.height - self._MAX_SCROLL_BAR_SIZE,
                                              self._MIN_SCREENSHOT_PART_HEIGHT)}

        logger.debug("Total size: {0}, Screenshot part size: {1}".format(entire_page_size,
                                                                         screenshot_part_size))

        entire_page = Region(0, 0, entire_page_size['width'], entire_page_size['height'])
        screenshot_parts = entire_page.get_sub_regions(screenshot_part_size)

        # Starting with the screenshot we already captured at (0,0).
        stitched_image = Image.new('RGBA', (entire_page.width, entire_page.height))
        stitched_image.paste(screenshot, box=(0, 0))
        self.save_position()

        for part in screenshot_parts:
            # Since we already took the screenshot for 0,0
            if part.left == 0 and part.top == 0:
                logger.debug('Skipping screenshot for 0,0 (already taken)')
                continue
            logger.debug("Taking screenshot for {0}".format(part))
            # Scroll to the part's top/left and give it time to stabilize.
            self.scroll_to(Point(part.left, part.top))
            EyesWebDriver._wait_before_screenshot(wait_before_screenshots)
            # Since screen size might cause the scroll to reach only part of the way
            current_scroll_position = self.get_current_position()
            logger.debug("Scrolled To ({0},{1})".format(current_scroll_position.x,
                                                        current_scroll_position.y))
            part64 = self.get_screenshot_as_base64()
            part_image = image_utils.image_from_bytes(base64.b64decode(part64))

            if need_to_scale:
                part_image = image_utils.scale_image(part_image, 1.0 / pixel_ratio)

            stitched_image.paste(part_image, box=(current_scroll_position.x, current_scroll_position.y))

        self.restore_position()
        self.restore_origin()
        self.switch_to.frames(original_frame)

        return stitched_image

    def get_stitched_screenshot(self, element_region, wait_before_screenshots, scale_provider):
        # type: (Region, int, ScaleProvider) -> Image.Image
        """
        Gets a stitched screenshot for specific element

        :param element_region: Region of required screenshot
        :param wait_before_screenshots: Seconds to wait before taking each screenshot.
        :param scale_provider: Scale image if needed.
        :return: The full element screenshot.
        """
        logger.info('getting stitched element screenshot..')
        self._position_provider = self._eyes._element_position_provider
        entire_size = self._position_provider.get_entire_size()
        logger.debug("Element region: {}".format(element_region))
        # Firefox 60 and above make a screenshot of the current frame when other browsers
        # make a screenshot of the viewport. So we scroll down to frame at _will_switch_to method
        # and add a left margin here.
        # TODO: Refactor code. Use EyesScreenshot
        if self._frame_chain:
            if ((self.browser_name == 'firefox' and self.browser_version < 60.0)
                    or self.browser_name in ('chrome', 'MicrosoftEdge', 'internet explorer', 'safari')):
                element_region.left += int(self._frame_chain.peek.location.x)

        screenshot_part_size = {'width':  element_region.width,
                                'height': max(element_region.height - self._MAX_SCROLL_BAR_SIZE,
                                              self._MIN_SCREENSHOT_PART_HEIGHT)}
        entire_element = Region(0, 0, entire_size['width'], entire_size['height'])

        screenshot_parts = entire_element.get_sub_regions(screenshot_part_size)
        viewport = self.get_viewport_size()
        screenshot = image_utils.image_from_bytes(base64.b64decode(self.get_screenshot_as_base64()))
        scale_provider.update_scale_ratio(screenshot.width)
        pixel_ratio = 1 / scale_provider.scale_ratio
        need_to_scale = True if pixel_ratio != 1.0 else False

        if need_to_scale:
            element_region = element_region.scale(scale_provider.device_pixel_ratio)

        # Starting with element region size part of the screenshot. Use it as a size template.
        stitched_image = Image.new('RGBA', (entire_element.width, entire_element.height))
        for part in screenshot_parts:
            logger.debug("Taking screenshot for {0}".format(part))
            # Scroll to the part's top/left and give it time to stabilize.
            self._position_provider.set_position(Point(part.left, part.top))
            EyesWebDriver._wait_before_screenshot(wait_before_screenshots)
            # Since screen size might cause the scroll to reach only part of the way
            current_scroll_position = self._position_provider.get_current_position()
            logger.debug("Scrolled To ({0},{1})".format(current_scroll_position.x,
                                                        current_scroll_position.y))
            part64 = self.get_screenshot_as_base64()
            part_image = image_utils.image_from_bytes(base64.b64decode(part64))
            # Cut to viewport size the full page screenshot of main frame for some browsers
            if self._frame_chain:
                if (self.browser_name == 'firefox' and self.browser_version < 60.0
                        or self.browser_name in ('internet explorer', 'safari')):
                    # TODO: Refactor this to make main screenshot only once
                    frame_scroll_position = int(self._frame_chain.peek.location.y)
                    part_image = image_utils.get_image_part(part_image, Region(top=frame_scroll_position,
                                                                               height=viewport['height'],
                                                                               width=viewport['width']))
            # We cut original image before scaling to prevent appearing of artifacts
            part_image = image_utils.get_image_part(part_image, element_region)
            if need_to_scale:
                part_image = image_utils.scale_image(part_image, 1.0 / pixel_ratio)

            # first iteration
            if stitched_image is None:
                stitched_image = part_image
                continue
            stitched_image.paste(part_image, box=(current_scroll_position.x, current_scroll_position.y))

        self._position_provider = self._origin_position_provider
        return stitched_image

    @property
    def switch_to(self):
        return _EyesSwitchTo(self, self.driver.switch_to)

    @property
    def current_offset(self):
        # type: () -> Point
        """
        Return the current offset of the context we're in (e.g., due to switching into frames)
        """
        x, y = 0, 0
        for frame in self._frame_chain:
            x += frame.location['x']
            y += frame.location['y']
        return Point(x, y)

    def execute_script(self, script, *args):
        return self.driver.execute_script(script, *args)

    def get_window_size(self, windowHandle='current'):
        return self.driver.get_window_size(windowHandle)

    def set_window_size(self, width, height, windowHandle='current'):
        self.driver.set_window_size(width, height, windowHandle)

    def set_window_position(self, x, y, windowHandle='current'):
        self.driver.set_window_position(x, y, windowHandle)



from .__version__ import __version__
from .core import *  # noqa
from .selenium import *  # noqa
from .utils import *  # noqa

# for backward compatibility
from .core import errors, geometry  # noqa
from .selenium import eyes, target  # noqa

__all__ = (
        core.__all__ +  # noqa
        selenium.__all__ +  # noqa
        utils.__all__ +  # noqa
        ('errors', 'geometry', 'target', 'StitchMode', 'eyes')
)

# for backward compatibility
VERSION = __version__



"""
General purpose utilities.
"""
from __future__ import absolute_import

import json
import time
import types
import itertools
import typing as tp
from datetime import timedelta, tzinfo

from applitools.core import logger
from .compat import urlparse, Queue

if tp.TYPE_CHECKING:
    from selenium.webdriver.remote.webdriver import WebDriver
    from selenium.webdriver.remote.webelement import WebElement
    from selenium.webdriver.remote.switch_to import SwitchTo

    from ..selenium.webdriver import EyesWebDriver, _EyesSwitchTo
    from ..selenium.webelement import EyesWebElement


class _UtcTz(tzinfo):
    """
    A UTC timezone class which is tzinfo compliant.
    """
    _ZERO = timedelta(0)

    def utcoffset(self, dt):
        return _UtcTz._ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return _UtcTz._ZERO


# Constant representing UTC
UTC = _UtcTz()


def to_json(obj):
    # type: (tp.Dict[str, tp.Any]) -> str
    """
    Returns an object's json representation (defaults to __getstate__ for user defined types).
    """
    return json.dumps(obj, default=lambda o: o.__getstate__(), indent=4)


def create_proxy_property(property_name, target_name, is_settable=False):
    # type: (str, str, bool) -> property
    """
    Returns a property object which forwards "name" to target.

    :param property_name: The name of the property.
    :param target_name: The target to forward to.
    """

    # noinspection PyUnusedLocal
    def _proxy_get(self):
        # type: (tp.Any) -> tp.Dict[str, float]
        return getattr(getattr(self, target_name), property_name)

    # noinspection PyUnusedLocal
    def _proxy_set(self, val):
        return setattr(getattr(self, target_name), property_name, val)

    if not is_settable:
        return property(_proxy_get)
    else:
        return property(_proxy_get, _proxy_set)


def create_forwarded_method(from_,  # type: tp.Union[EyesWebDriver, EyesWebElement, _EyesSwitchTo]
                            to,  # type: tp.Union[WebDriver, WebElement, SwitchTo]
                            func_name,  # type: str
                            ):
    # type: (...) -> tp.Callable
    """
    Returns a method(!) to be set on 'from_', which activates 'func_name' on 'to'.

    :param from_: Source.
    :param to: Destination.
    :param func_name: The name of function to activate.
    :return: Relevant method.
    """

    # noinspection PyUnusedLocal
    def forwarded_method(self_, *args, **kwargs):
        # type: (EyesWebDriver, *tp.Any, **tp.Any) -> tp.Callable
        return getattr(to, func_name)(*args, **kwargs)

    return types.MethodType(forwarded_method, from_)


def create_proxy_interface(from_,  # type: tp.Union[EyesWebDriver, EyesWebElement, _EyesSwitchTo]
                           to,  # type: tp.Union[WebDriver, WebElement, SwitchTo]
                           ignore_list=None,  # type: tp.List[str]
                           override_existing=False,  # type: bool
                           ):
    # type: (...) -> None
    """
    Copies the public interface of the destination object, excluding names in the ignore_list,
    and creates an identical interface in 'src', which forwards calls to dst.

    :param from_: Source.
    :param to: Destination.
    :param ignore_list: List of names to ignore while copying.
    :param override_existing: If False, attributes already existing in 'src' will not be overridden.
    """
    if not ignore_list:
        ignore_list = []
    for attr_name in dir(to):
        if not attr_name.startswith('_') and attr_name not in ignore_list:
            if callable(getattr(to, attr_name)):
                if override_existing or not hasattr(from_, attr_name):
                    setattr(from_, attr_name, create_forwarded_method(from_, to, attr_name))


def cached_property(f):
    # type: (tp.Callable) -> tp.Any
    """
    Returns a cached property that is calculated by function f
    """

    def get(self):
        try:
            return self._property_cache[f]
        except AttributeError:
            self._property_cache = {}
            x = self._property_cache[f] = f(self)
            return x
        except KeyError:
            x = self._property_cache[f] = f(self)
            return x

    return property(get)


def is_absolute_url(url):
    return bool(urlparse(url).netloc)


def is_url_with_scheme(url):
    return bool(urlparse(url).scheme)


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            logger.debug('%r  %2.2f ms' % (method.__name__, (te - ts) * 1000))
        return result

    return timed


def retry(delays=(0, 1, 5),
          exception=Exception,
          report=lambda *args: None):
    """
    This is a Python decorator which helps implementing an aspect oriented
    implementation of a retrying of certain steps which might fail sometimes.
    https://code.activestate.com/recipes/580745-retry-decorator-in-python/
    """

    def wrapper(function):
        def wrapped(*args, **kwargs):
            problems = []
            for delay in itertools.chain(delays, [None]):
                try:
                    return function(*args, **kwargs)
                except exception as problem:
                    problems.append(problem)
                    if delay is None:
                        report("retryable failed definitely:", problems)
                        raise
                    else:
                        report("retryable failed:", problem,
                               "-- delaying for %ds" % delay)
                        time.sleep(delay)

        return wrapped

    return wrapper



from .capture import EyesWebDriverScreenshot, dom_capture
from .positioning import StitchMode
from .eyes import Eyes
from .webdriver import EyesWebDriver
from .webelement import EyesWebElement
from .target import (IgnoreRegionByElement, IgnoreRegionBySelector, FloatingBounds, FloatingRegion,
                     FloatingRegionByElement, FloatingRegionBySelector, Target)
from .frames import Frame

__all__ = (
        target.__all__ +  # noqa
        ('Eyes', 'EyesWebElement', 'EyesWebDriver', 'Frame', 'EyesWebDriverScreenshot',
         'StitchMode', 'dom_capture'))



from __future__ import absolute_import

import time
import typing as tp
from contextlib import contextmanager

from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webdriver import WebDriver

from ..core import logger, EyesError

if tp.TYPE_CHECKING:
    from applitools.utils.custom_types import AnyWebDriver, ViewPort

__all__ = ('get_current_frame_content_entire_size', 'get_device_pixel_ratio', 'get_viewport_size',
           'get_window_size', 'set_window_size', 'set_browser_size', 'set_browser_size_by_viewport_size',
           'set_viewport_size', 'hide_scrollbars', 'set_overflow')

_NATIVE_APP = 'NATIVE_APP'
_JS_GET_VIEWPORT_SIZE = """
    var height = undefined;
    var width = undefined;
    if (window.innerHeight) {
        height = window.innerHeight;
    } else if (document.documentElement && document.documentElement.clientHeight) {
        height = document.documentElement.clientHeight;
    } else {
            var b = document.getElementsByTagName('body')[0];
            if (b.clientHeight) {height = b.clientHeight;}
        }
    if (window.innerWidth) {
        width = window.innerWidth;
    } else if (document.documentElement && document.documentElement.clientWidth) {
        width = document.documentElement.clientWidth;
    } else {
        var b = document.getElementsByTagName('body')[0];
    if (b.clientWidth) {
        width = b.clientWidth;}
    }
    return [width, height];"""

_JS_GET_CURRENT_SCROLL_POSITION = """
    var doc = document.documentElement;
    var x = window.scrollX || ((window.pageXOffset || doc.scrollLeft) - (doc.clientLeft || 0));
    var y = window.scrollY || ((window.pageYOffset || doc.scrollTop) - (doc.clientTop || 0));
    return [x, y]"""

_JS_GET_CONTENT_ENTIRE_SIZE = """
    var scrollWidth = document.documentElement.scrollWidth;
    var bodyScrollWidth = document.body.scrollWidth;
    var totalWidth = Math.max(scrollWidth, bodyScrollWidth);
    var clientHeight = document.documentElement.clientHeight;
    var bodyClientHeight = document.body.clientHeight;
    var scrollHeight = document.documentElement.scrollHeight;
    var bodyScrollHeight = document.body.scrollHeight;
    var maxDocElementHeight = Math.max(clientHeight, scrollHeight);
    var maxBodyHeight = Math.max(bodyClientHeight, bodyScrollHeight);
    var totalHeight = Math.max(maxDocElementHeight, maxBodyHeight);
    return [totalWidth, totalHeight];";
"""
_JS_SET_OVERFLOW = """
  return (function() {
    var origOF = document.documentElement.style.overflow;
    document.documentElement.style.overflow = '{overflow}';
    return origOF;
  }());
"""

_JS_TRANSFORM_KEYS = ("transform", "-webkit-transform")
_OVERFLOW_HIDDEN = 'hidden'
_MAX_DIFF = 3
_SLEEP = 1  # sec
_RETRIES = 3


def is_mobile_device(driver):
    # type: (AnyWebDriver) -> bool
    """
    Returns whether the platform running is a mobile device or not.

    :return: True if the platform running the test is a mobile platform. False otherwise.
    """
    is_mobile = """
if( navigator.userAgent.match(/Android/i) ||
    navigator.userAgent.match(/iPhone/i) ||
    navigator.userAgent.match(/iPad/i)   ||
    navigator.userAgent.match(/iPod/i) ) {
    return true;
} else {
    return false;
}
    """
    # TODO: Implement proper UserAgent handling
    driver = get_underlying_driver(driver)
    is_mobile_platform = driver.desired_capabilities.get('platformName') in ('Android', 'iOS')
    if not is_mobile_platform:
        is_mobile_platform = driver.execute_script(is_mobile)
    return is_mobile_platform


def get_underlying_driver(driver):
    # type: (AnyWebDriver) -> WebDriver
    from ..selenium.webdriver import EyesWebDriver
    if isinstance(driver, EyesWebDriver):
        driver = driver.driver
    return driver


def get_current_frame_content_entire_size(driver):
    # type: (AnyWebDriver) -> ViewPort
    """
    :return: The size of the entire content.
    """
    try:
        width, height = driver.execute_script(_JS_GET_CONTENT_ENTIRE_SIZE)
    except WebDriverException:
        raise EyesError('Failed to extract entire size!')
    return dict(width=width, height=height)


def get_device_pixel_ratio(driver):
    # type: (AnyWebDriver) -> float
    return driver.execute_script('return window.devicePixelRatio;')


def get_viewport_size(driver):
    # type: (AnyWebDriver) -> ViewPort
    """
    Tries to get the viewport size using Javascript. If fails, gets the entire browser window
    size!

    :param driver: The webdriver to use for getting the viewport size.
    """
    # noinspection PyBroadException
    try:
        width, height = driver.execute_script(_JS_GET_VIEWPORT_SIZE)
        return dict(width=width, height=height)
    except WebDriverException:
        logger.info('Failed to get viewport size. Only window size is available')
        return get_window_size(driver)


def get_window_size(driver):
    # type: (AnyWebDriver) -> ViewPort
    return driver.get_window_size()


def set_window_size(driver, size):
    # type: (AnyWebDriver, ViewPort) -> None
    driver.set_window_size(size['width'], size['height'])


def set_browser_size(driver, required_size):
    # type: (AnyWebDriver, ViewPort) -> bool

    retries_left = _RETRIES

    # set browser size for mobile devices isn't working
    if is_mobile_device(driver):
        return True

    while True:
        logger.info("Trying to set browser size to: " + str(required_size))
        set_window_size(driver, required_size)
        time.sleep(_SLEEP)
        current_size = get_window_size(driver)
        logger.info("Current browser size: " + str(required_size))

        if retries_left or current_size != required_size:
            break
        retries_left -= 1

    return current_size == required_size


def set_browser_size_by_viewport_size(driver, actual_viewport_size, required_size):
    # type: (AnyWebDriver, ViewPort, ViewPort) -> bool

    browser_size = get_window_size(driver)
    logger.debug("Current browser size: {}".format(browser_size))
    required_browser_size = dict(
            width=browser_size['width'] + (required_size['width'] - actual_viewport_size['width']),
            height=browser_size['height'] + (required_size['height'] - actual_viewport_size['height'])
    )
    return set_browser_size(driver, required_browser_size)


def set_viewport_size(driver, required_size):
    # type: (AnyWebDriver, ViewPort) -> None

    logger.info("set_viewport_size({})".format(str(required_size)))

    actual_viewport_size = get_viewport_size(driver)
    if actual_viewport_size == required_size:
        logger.info("Required size already set.")
        return None

    try:
        # We move the window to (0,0) to have the best chance to be able to
        # set the viewport size as requested.
        driver.set_window_position(0, 0)
    except WebDriverException:
        logger.info('Warning: Failed to move the browser window to (0,0)')
    set_browser_size_by_viewport_size(driver, actual_viewport_size, required_size)
    actual_viewport_size = get_viewport_size(driver)
    if actual_viewport_size == required_size:
        return None

    # Additional attempt. This Solves the "maximized browser" bug
    # (border size for maximized browser sometimes different than
    # non-maximized, so the original browser size calculation is
    # wrong).
    logger.info("Trying workaround for maximization...")
    set_browser_size_by_viewport_size(driver, actual_viewport_size, required_size)
    actual_viewport_size = get_viewport_size(driver)
    logger.debug("Current viewport size: {}".format(actual_viewport_size))
    if actual_viewport_size == required_size:
        return None

    width_diff = abs(actual_viewport_size['width'] - required_size['width'])
    width_step = -1 if width_diff > 0 else 1  # -1 for smaller size, 1 for larger
    height_diff = abs(actual_viewport_size['height'] - required_size['height'])
    height_step = -1 if height_diff > 0 else 1

    browser_size = get_window_size(driver)
    curr_width_change = 0
    curr_height_change = 0

    if width_diff <= _MAX_DIFF and height_diff <= _MAX_DIFF:
        logger.info("Trying workaround for zoom...")
        last_required_browser_size = None

        while abs(curr_width_change) <= width_diff and abs(curr_height_change) <= height_diff:
            if abs(curr_width_change) <= width_diff:
                curr_width_change += width_step
            if abs(curr_height_change) <= height_diff:
                curr_height_change += height_step

            required_browser_size = dict(
                    width=browser_size['width'] + curr_width_change,
                    height=browser_size['height'] + curr_height_change)
            if required_browser_size == last_required_browser_size:
                logger.info("Browser size is as required but viewport size does not match!")
                logger.info("Browser size: {}, Viewport size: {}".format(required_browser_size,
                                                                         actual_viewport_size))
                logger.info("Stopping viewport size attempts.")
                break

            set_browser_size(driver, required_browser_size)
            last_required_browser_size = required_browser_size

            actual_viewport_size = get_viewport_size(driver)
            logger.info("Current viewport size: {}".format(actual_viewport_size))

            if actual_viewport_size == required_size:
                return None
        else:
            logger.info('Zoom workaround failed.')
    raise EyesError('Failed to set the viewport size.')


def hide_scrollbars(driver):
    return set_overflow(driver, _OVERFLOW_HIDDEN)


def set_overflow(driver, overflow):
    with timeout(0.1):
        return driver.execute_script(_JS_SET_OVERFLOW.format(overflow=overflow))


@contextmanager
def timeout(timeout):
    time.sleep(timeout)
    yield


def is_landscape_orientation(driver):
    if is_mobile_device(driver):
        appium_driver = get_underlying_driver(driver)  # type: AppiumRemoteWebDriver

        original_context = None
        try:
            # We must be in native context in order to ask for orientation,
            # because of an Appium bug.
            original_context = appium_driver.context
            if len(appium_driver.contexts) > 1 and not original_context.uppar() == 'NATIVE_APP':
                appium_driver.switch_to.context('NATIVE_APP')
            else:
                original_context = None
        except WebDriverException as e:
            original_context = None

        try:
            orieintation = appium_driver.orientation
            return orieintation.lower() == 'landscape'
        except Exception as e:
            logger.debug("WARNING: Couldn't get device orientation. Assuming Portrait.")
        finally:
            if original_context is not None:
                appium_driver.switch_to.context(original_context)

        return False


def get_viewport_size_or_display_size(driver):
    logger.info("get_viewport_size_or_display_size()")

    try:
        return get_viewport_size(driver)
    except Exception as e:
        logger.info("Failed to extract viewport size using Javascript: {}".format(str(e)))
    # If we failed to extract the viewport size using JS, will use the
    # window size instead.

    logger.info("Using window size as viewport size.")
    window_size = get_window_size(driver)
    width = window_size['width']
    height = window_size['height']
    try:
        if is_landscape_orientation(driver) and height > width:
            height, width = width, height
    except WebDriverException as e:
        # Not every WebDriver supports querying for orientation.
        pass
    logger.info("Done! Size {:d} x {:d}".format(width, height))
    return dict(width=width, height=height)



from __future__ import absolute_import

import math
import time
import typing as tp

from selenium.webdriver.common.by import By

from applitools.core.geometry import Region
from applitools.core import logger
from applitools.utils import general_utils

if tp.TYPE_CHECKING:
    from selenium.webdriver.remote.webelement import WebElement

    from applitools.core.geometry import Point
    from applitools.utils.custom_types import AnyWebDriver
    from .webdriver import EyesWebDriver
    from .eyes import Eyes


class EyesWebElement(object):
    """
    A wrapper for selenium web element. This enables eyes to be notified about actions/events for
    this element.
    """
    _METHODS_TO_REPLACE = ['find_element', 'find_elements']

    # Properties require special handling since even testing if they're callable "activates"
    # them, which makes copying them automatically a problem.
    _READONLY_PROPERTIES = ['tag_name', 'text', 'location_once_scrolled_into_view',
                            'parent', 'rect', 'screenshot_as_base64', 'screenshot_as_png',
                            'location_in_view', 'anonymous_children']
    _JS_GET_COMPUTED_STYLE_FORMATTED_STR = """
            var elem = arguments[0];
            var styleProp = '%s';
            if (window.getComputedStyle) {
                return window.getComputedStyle(elem, null)
                .getPropertyValue(styleProp);
            } else if (elem.currentStyle) {
                return elem.currentStyle[styleProp];
            } else {
                return null;
            }
    """
    _JS_GET_SCROLL_LEFT = "return arguments[0].scrollLeft;"
    _JS_GET_SCROLL_TOP = "return arguments[0].scrollTop;"
    _JS_GET_SCROLL_WIDTH = "return arguments[0].scrollWidth;"
    _JS_GET_SCROLL_HEIGHT = "return arguments[0].scrollHeight;"
    _JS_GET_OVERFLOW = "return arguments[0].style.overflow;"
    _JS_SET_OVERFLOW_FORMATTED_STR = "arguments[0].style.overflow = '%s'"
    _JS_GET_CLIENT_WIDTH = "return arguments[0].clientWidth;"
    _JS_GET_CLIENT_HEIGHT = "return arguments[0].clientHeight;"
    _JS_SCROLL_TO_FORMATTED_STR = """
            arguments[0].scrollLeft = {:d};
            arguments[0].scrollTop = {:d};
    """
    _JS_GET_BORDER_WIDTHS_ARR = """
            var retVal = retVal || [];
            if (window.getComputedStyle) {
            var computedStyle = window.getComputedStyle(elem, null);
            retVal.push(computedStyle.getPropertyValue('border-left-width'));
            retVal.push(computedStyle.getPropertyValue('border-top-width'));
            retVal.push(computedStyle.getPropertyValue('border-right-width'));
            retVal.push(computedStyle.getPropertyValue('border-bottom-width'));
            } else if (elem.currentStyle) {
            retVal.push(elem.currentStyle['border-left-width']);
            retVal.push(elem.currentStyle['border-top-width']);
            retVal.push(elem.currentStyle['border-right-width']);
            retVal.push(elem.currentStyle['border-bottom-width']);
            } else {
            retVal.push(0,0,0,0);
            }
    """
    _JS_GET_BORDER_WIDTHS = _JS_GET_BORDER_WIDTHS_ARR + "return retVal;"
    _JS_GET_SIZE_AND_BORDER_WIDTHS = """
            var elem = arguments[0];
            var retVal = [arguments[0].clientWidth, arguments[0].clientHeight];
            {}
            return retVal
    """.format(_JS_GET_BORDER_WIDTHS_ARR)

    def __init__(self, element, driver):
        # type: (WebElement, EyesWebDriver) -> None
        """
        Ctor.

        :param element: The element in the frame.
        :param eyes: The eyes sdk instance.
        :param driver: EyesWebDriver instance.
        """
        self.element = element
        self._driver = driver  # type: AnyWebDriver
        # Replacing implementation of the underlying driver with ours. We'll put the original
        # methods back before destruction.
        self._original_methods = {}  # type: tp.Dict[tp.Text, tp.Callable]
        for method_name in self._METHODS_TO_REPLACE:
            self._original_methods[method_name] = getattr(element, method_name)
            setattr(element, method_name, getattr(self, method_name))

        # Copies the web element's interface
        general_utils.create_proxy_interface(self, element, self._READONLY_PROPERTIES)
        # Setting properties
        for attr in self._READONLY_PROPERTIES:
            setattr(self.__class__, attr, general_utils.create_proxy_property(attr, 'element'))

    @property
    def bounds(self):
        # type: () -> Region
        # noinspection PyUnresolvedReferences
        location = self.location
        left, top = location['x'], location['y']
        width = height = 0  # Default

        # noinspection PyBroadException
        try:
            size = self.element.size
            width, height = size['width'], size['height']
        except Exception:
            # Not implemented on all platforms.
            pass
        if left < 0:
            left, width = 0, max(0, width + left)
        if top < 0:
            top, height = 0, max(0, height + top)
        return Region(left, top, width, height)

    @property
    def location(self):
        return self.element.location

    @property
    def size(self):
        return self.element.size

    @property
    def id(self):
        return self.element.id

    def value_of_css_property(self, property_name):
        return self.element.value_of_css_property(property_name)

    def find_element(self, by=By.ID, value=None):
        """
        Returns a WebElement denoted by "By".

        :param by: By which option to search for (default is by ID).
        :param value: The value to search for.
        :return: WebElement denoted by "By".
        """
        # Get result from the original implementation of the underlying driver.
        result = self._original_methods['find_element'](by, value)
        # Wrap the element.
        if result:
            result = EyesWebElement(result, self._driver)
        return result

    def find_elements(self, by=By.ID, value=None):
        """
        Returns a list of web elements denoted by "By".

        :param by: By which option to search for (default is by ID).
        :param value: The value to search for.
        :return: List of web elements denoted by "By".
        """
        # Get result from the original implementation of the underlying driver.
        results = self._original_methods['find_elements'](by, value)
        # Wrap all returned elements.
        if results:
            updated_results = []
            for element in results:
                updated_results.append(EyesWebElement(element, self._driver))
            results = updated_results
        return results

    def click(self):
        """
        Clicks and element.
        """
        self._driver._eyes.add_mouse_trigger_by_element('click', self)
        self.element.click()

    def send_keys(self, *value):
        """
        Sends keys to a certain element.

        :param value: The value to type into the element.
        """
        text = u''
        for val in value:
            if isinstance(val, int):
                val = val.__str__()
            text += val.encode('utf-8').decode('utf-8')
        self._driver._eyes.add_text_trigger_by_element(self, text)
        self.element.send_keys(*value)

    def set_overflow(self, overflow, stabilization_time=None):
        """
        Sets the overflow of the current element.

        :param overflow: The overflow value to set. If the given value is None, then overflow will be set to
                         undefined.
        :param stabilization_time: The time to wait for the page to stabilize after overflow is set. If the value is
                                    None, then no waiting will take place. (Milliseconds)
        :return: The previous overflow value.
        """
        logger.debug("Setting overflow: %s" % overflow)
        if overflow is None:
            script = "var elem = arguments[0]; var origOverflow = elem.style.overflow; " \
                     "elem.style.overflow = undefined; " \
                     "return origOverflow;"
        else:
            script = "var elem = arguments[0]; var origOverflow = elem.style.overflow; " \
                     "elem.style.overflow = \"{0}\"; " \
                     "return origOverflow;".format(overflow)
        # noinspection PyUnresolvedReferences
        original_overflow = self._driver.execute_script(script, self.element)
        logger.debug("Original overflow: %s" % original_overflow)
        if stabilization_time is not None:
            time.sleep(stabilization_time / 1000)
        return original_overflow

    def hide_scrollbars(self):
        # type: () -> tp.Text
        """
        Hides the scrollbars of the current element.

        :return: The previous value of the overflow property (could be None).
        """
        logger.debug('EyesWebElement.HideScrollbars()')
        return self.set_overflow('hidden')

    def get_computed_style(self, prop_style):
        script = self._JS_GET_COMPUTED_STYLE_FORMATTED_STR % prop_style
        return self._driver.execute_script(script, self.element)

    def get_computed_style_int(self, prop_style):

        value = self.get_computed_style(prop_style)
        return int(round(float(value.replace('px', '').strip())))

    def get_scroll_left(self):
        return int(math.ceil(self._driver.execute_script(self._JS_GET_SCROLL_LEFT, self.element)))

    def get_scroll_top(self):
        return math.ceil(self._driver.execute_script(self._JS_GET_SCROLL_TOP, self.element))

    def get_scroll_width(self):
        return int(math.ceil(self._driver.execute_script(self._JS_GET_SCROLL_WIDTH, self.element)))

    def get_scroll_height(self):
        return int(math.ceil(self._driver.execute_script(self._JS_GET_SCROLL_HEIGHT, self.element)))

    def get_border_left_width(self):
        return self.get_computed_style_int('border-left-width')

    def get_border_right_width(self):
        return self.get_computed_style_int('border-right-width')

    def get_border_top_width(self):
        return self.get_computed_style_int('border-right-width')

    def get_border_bottom_width(self):
        return self.get_computed_style_int('border-right-width')

    def get_overflow(self):
        return self._driver.execute_script(self._JS_GET_OVERFLOW, self.element)

    def get_client_width(self):
        return int(math.ceil(float(self._driver.execute_script(self._JS_GET_CLIENT_WIDTH, self.element))))

    def get_client_height(self):
        return int(math.ceil(float(self._driver.execute_script(self._JS_GET_CLIENT_HEIGHT, self.element))))

    def scroll_to(self, location):
        # type: (Point) -> None
        """Scrolls to the specified location inside the element."""
        self._driver.execute_script(
            self._JS_SCROLL_TO_FORMATTED_STR.format(location.x, location.y), self.element)

    @property
    def size_and_borders(self):
        ret_val = self._driver.execute_script(self._JS_GET_SIZE_AND_BORDER_WIDTHS, self.element)
        return SizeAndBorders(
            width=int(ret_val[0]),
            height=int(ret_val[1]),
            left=int(ret_val[2].rstrip('px')),
            top=int(ret_val[3].rstrip('px')),
            right=int(ret_val[4].rstrip('px')),
            bottom=int(ret_val[5].rstrip('px'))
        )

    def __str__(self):
        return "EyesWebElement: id {}, tag_name {}".format(self.element.id, self.element.tag_name, )


class SizeAndBorders(object):
    __slots__ = ('size', 'borders')

    def __init__(self, width, height, left, top, right, bottom):
        self.size = dict(width=width, height=height)
        self.borders = dict(left=left, top=top, right=right, bottom=bottom)



from __future__ import absolute_import

import base64
import contextlib
import typing as tp

from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver

# noinspection PyProtectedMember
from applitools.core import logger
from applitools.core.eyes_base import FailureReports, EyesBase
from applitools.core.match_window_task import MatchWindowTask
from applitools.core.triggers import MouseTrigger, TextTrigger
from applitools.core.errors import EyesError, TestFailedError
from applitools.core.geometry import Region
from applitools.core.scaling import ContextBasedScaleProvider, FixedScaleProvider
from applitools.utils import image_utils
from . import eyes_selenium_utils
from .webdriver import EyesWebDriver
from .capture import EyesWebDriverScreenshot, dom_capture
from .target import Target
from .positioning import StitchMode, ElementPositionProvider

if tp.TYPE_CHECKING:
    from applitools.core.scaling import ScaleProvider
    from applitools.utils.custom_types import (ViewPort, AnyWebDriver, FrameReference, AnyWebElement)


class ScreenshotType(object):
    ENTIRE_ELEMENT_SCREENSHOT = 'EntireElementScreenshot'
    REGION_OR_ELEMENT_SCREENSHOT = 'RegionOrElementScreenshot'
    FULLPAGE_SCREENSHOT = "FullPageScreenshot"
    VIEWPORT_SCREENSHOT = "ViewportScreenshot"


class Eyes(EyesBase):
    """
    Applitools Selenium Eyes API for python.
    """
    _UNKNOWN_DEVICE_PIXEL_RATIO = 0
    _DEFAULT_DEVICE_PIXEL_RATIO = 1

    @staticmethod
    def set_viewport_size(driver, size):
        # type: (AnyWebDriver, ViewPort) -> None
        assert driver is not None
        eyes_selenium_utils.set_viewport_size(driver, size)

    def get_viewport_size(self):
        # type: () -> ViewPort
        """
        Returns the size of the viewport of the application under test (e.g, the browser).
        """
        if self._viewport_size:
            return self._viewport_size
        self._viewport_size = self._driver.get_viewport_size()
        return self._viewport_size

    def __init__(self, server_url=EyesBase.DEFAULT_EYES_SERVER):
        super(Eyes, self).__init__(server_url)

        self._driver = None  # type: tp.Optional[AnyWebDriver]
        self._match_window_task = None  # type: tp.Optional[MatchWindowTask]
        self._viewport_size = None  # type: tp.Optional[ViewPort]
        self._screenshot_type = None  # type: tp.Optional[str]  # ScreenshotType
        self._device_pixel_ratio = self._UNKNOWN_DEVICE_PIXEL_RATIO
        self._stitch_mode = StitchMode.Scroll  # type: tp.Text
        self._element_position_provider = None  # type: tp.Optional[ElementPositionProvider]

        # If true, Eyes will create a full page screenshot (by using stitching) for browsers which only
        # returns the viewport screenshot.
        self.force_full_page_screenshot = False  # type: bool

        # If true, Eyes will remove the scrollbars from the pages before taking the screenshot.
        self.hide_scrollbars = False  # type: bool

    @property
    def stitch_mode(self):
        # type: () -> tp.Text
        """
        Gets the stitch mode.

        :return: The stitch mode.
        """
        return self._stitch_mode

    @stitch_mode.setter
    def stitch_mode(self, stitch_mode):
        # type: (tp.Text) -> None
        """
        Sets the stitch property - default is by scrolling.

        :param stitch_mode: The stitch mode to set - either scrolling or css.
        """
        self._stitch_mode = stitch_mode
        if stitch_mode == StitchMode.CSS:
            self.hide_scrollbars = True
            self.send_dom = True

    @property
    def driver(self):
        # type: () -> EyesWebDriver
        """
        Returns the current web driver.
        """
        return self._driver

    def _obtain_screenshot_type(self, is_element, inside_a_frame, stitch_content, force_fullpage, is_region=False):
        # type:(bool, bool, bool, bool, bool) -> str
        if stitch_content or force_fullpage:
            if is_element and not stitch_content:
                return ScreenshotType.REGION_OR_ELEMENT_SCREENSHOT

            if not inside_a_frame:
                if ((force_fullpage and not stitch_content) or  # noqa
                        (stitch_content and not is_element)):
                    return ScreenshotType.FULLPAGE_SCREENSHOT

            if inside_a_frame or stitch_content:
                return ScreenshotType.ENTIRE_ELEMENT_SCREENSHOT

        else:
            if is_region or (is_element and not stitch_content):
                return ScreenshotType.REGION_OR_ELEMENT_SCREENSHOT

            if not stitch_content and not force_fullpage:
                return ScreenshotType.VIEWPORT_SCREENSHOT

        return ScreenshotType.VIEWPORT_SCREENSHOT

    @property
    def _environment(self):
        os = self.host_os
        # If no host OS was set, check for mobile OS.
        if os is None:
            logger.info('No OS set, checking for mobile OS...')
            # Since in Python Appium driver is the same for Android and iOS, we need to use the desired
            # capabilities to figure this out.
            if eyes_selenium_utils.is_mobile_device(self._driver):
                platform_name = self._driver.platform_name
                logger.info(platform_name + ' detected')
                platform_version = self._driver.platform_version
                if platform_version is not None:
                    # Notice that Python's "split" function's +limit+ is the the maximum splits performed
                    # whereas in Ruby it is the maximum number of elements in the result (which is why they are set
                    # differently).
                    major_version = platform_version.split('.', 1)[0]
                    os = platform_name + ' ' + major_version
                else:
                    os = platform_name
                logger.info("Setting OS: " + os)
            else:
                logger.info('No mobile OS detected.')
        app_env = {'os':          os, 'hostingApp': self.host_app,
                   'displaySize': self._viewport_size,
                   'inferred':    self._inferred_environment}
        return app_env

    @property
    def _title(self):
        if self._should_get_title:
            # noinspection PyBroadException
            try:
                return self._driver.title
            except Exception:
                self._should_get_title = False
                # Couldn't get title, return empty string.
        return ''

    @property
    def _inferred_environment(self):
        # type: () -> tp.Optional[tp.Text]
        try:
            user_agent = self._driver.execute_script('return navigator.userAgent')
        except WebDriverException:
            user_agent = None
        if user_agent:
            return "useragent:%s" % user_agent
        return None

    def _update_scaling_params(self):
        # type: () -> tp.Optional[ScaleProvider]
        if self._device_pixel_ratio != self._UNKNOWN_DEVICE_PIXEL_RATIO:
            logger.debug("Device pixel ratio was already changed")
            return None

        logger.info('Trying to extract device pixel ratio...')
        try:
            device_pixel_ratio = image_utils.get_device_pixel_ratio(self._driver)
        except Exception as e:
            logger.info('Failed to extract device pixel ratio! Using default. Error %s ' % e)
            device_pixel_ratio = self._DEFAULT_DEVICE_PIXEL_RATIO
        logger.info('Device pixel ratio: {}'.format(device_pixel_ratio))

        logger.info("Setting scale provider...")
        try:
            scale_provider = ContextBasedScaleProvider(
                top_level_context_entire_size=self._driver.get_entire_page_size(),
                viewport_size=self.get_viewport_size(),
                device_pixel_ratio=device_pixel_ratio,
                # always False as in Java version
                is_mobile_device=False)  # type: ScaleProvider
        except Exception:
            # This can happen in Appium for example.
            logger.info("Failed to set ContextBasedScaleProvider.")
            logger.info("Using FixedScaleProvider instead...")
            scale_provider = FixedScaleProvider(1 / device_pixel_ratio)
        logger.info("Done!")
        return scale_provider

    @contextlib.contextmanager
    def _hide_scrollbars_if_needed(self):
        if self.hide_scrollbars:
            original_overflow = self._driver.hide_scrollbars()
        yield
        if self.hide_scrollbars:
            self._driver.set_overflow(original_overflow)

    def _try_capture_dom(self):
        try:
            dom_json = dom_capture.get_full_window_dom(self._driver)
            return dom_json
        except Exception as e:
            logger.warning(
                'Exception raising during capturing DOM Json. Passing...\n Got next error: {}'.format(str(e)))
            return None

    def _get_screenshot(self):
        scale_provider = self._update_scaling_params()

        if self._screenshot_type == ScreenshotType.ENTIRE_ELEMENT_SCREENSHOT:
            self._last_screenshot = self._entire_element_screenshot(scale_provider)

        elif self._screenshot_type == ScreenshotType.FULLPAGE_SCREENSHOT:
            self._last_screenshot = self._full_page_screenshot(scale_provider)

        elif self._screenshot_type == ScreenshotType.VIEWPORT_SCREENSHOT:
            self._last_screenshot = self._viewport_screenshot(scale_provider)

        elif self._screenshot_type == ScreenshotType.REGION_OR_ELEMENT_SCREENSHOT:
            self._last_screenshot = self._region_or_screenshot(scale_provider)

        else:
            raise EyesError("No proper ScreenshotType obtained")
        return self._last_screenshot

    def get_screenshot(self, hide_scrollbars_called=False):
        if hide_scrollbars_called:
            return self._get_screenshot()
        else:
            with self._hide_scrollbars_if_needed():
                return self._get_screenshot()

    def _entire_element_screenshot(self, scale_provider):
        # type: (ScaleProvider) -> EyesWebDriverScreenshot
        logger.info('Entire element screenshot requested')
        screenshot = self._driver.get_stitched_screenshot(self._region_to_check,
                                                          self._seconds_to_wait_screenshot,
                                                          scale_provider)
        return EyesWebDriverScreenshot.create_from_image(screenshot, self._driver)

    def _region_or_screenshot(self, scale_provider):
        # type: (ScaleProvider) -> EyesWebDriverScreenshot
        logger.info('Not entire element screenshot requested')
        screenshot = self._viewport_screenshot(scale_provider)
        region = screenshot.get_element_region_in_frame_viewport(self._region_to_check)
        screenshot = screenshot.get_sub_screenshot_by_region(region)
        return screenshot

    def _full_page_screenshot(self, scale_provider):
        # type: (ScaleProvider) -> EyesWebDriverScreenshot
        logger.info('Full page screenshot requested')
        screenshot = self._driver.get_full_page_screenshot(self._seconds_to_wait_screenshot,
                                                           scale_provider)
        return EyesWebDriverScreenshot.create_from_image(screenshot, self._driver)

    def _viewport_screenshot(self, scale_provider):
        # type: (ScaleProvider) -> EyesWebDriverScreenshot
        logger.info('Viewport screenshot requested')

        self._driver._wait_before_screenshot(self._seconds_to_wait_screenshot)
        if not self._driver.is_mobile_device():
            image64 = self._driver.get_screesnhot_as_base64_from_main_frame()
        else:
            image64 = self._driver.get_screenshot_as_base64()

        image = image_utils.image_from_bytes(base64.b64decode(image64))
        scale_provider.update_scale_ratio(image.width)
        pixel_ratio = 1 / scale_provider.scale_ratio
        if pixel_ratio != 1.0:
            image = image_utils.scale_image(image, 1.0 / pixel_ratio)
        return EyesWebDriverScreenshot.create_from_image(image, self._driver).get_viewport_screenshot()

    def _ensure_viewport_size(self):
        if self._viewport_size is None:
            self._viewport_size = self._driver.get_default_content_viewport_size()
            if not eyes_selenium_utils.is_mobile_device(self._driver):
                eyes_selenium_utils.set_viewport_size(self._driver, self._viewport_size)

    def open(self, driver, app_name, test_name, viewport_size=None):
        # type: (AnyWebDriver, tp.Text, tp.Text, tp.Optional[ViewPort]) -> EyesWebDriver
        if self.is_disabled:
            logger.debug('open(): ignored (disabled)')
            return driver

        if isinstance(driver, EyesWebDriver):
            # If the driver is an EyesWebDriver (as might be the case when tests are ran
            # consecutively using the same driver object)
            self._driver = driver
        else:
            if not isinstance(driver, RemoteWebDriver):
                logger.info("WARNING: driver is not a RemoteWebDriver (class: {0})".format(driver.__class__))
            self._driver = EyesWebDriver(driver, self, self._stitch_mode)

        if viewport_size is not None:
            self._viewport_size = viewport_size
            eyes_selenium_utils.set_viewport_size(self._driver, viewport_size)

        self._ensure_viewport_size()
        self._open_base(app_name, test_name, viewport_size)

        return self._driver

    def check_window(self, tag=None, match_timeout=-1, target=None):
        # type: (tp.Optional[tp.Text], int, tp.Optional[Target]) -> None
        """
        Takes a snapshot from the browser using the web driver and matches it with the expected
        output.

        :param tag: (str) Description of the visual validation checkpoint.
        :param match_timeout: (int) Timeout for the visual validation checkpoint (milliseconds).
        :param target: (Target) The target for the check_window call
        :return: None
        """
        logger.info("check_window('%s')" % tag)
        self._screenshot_type = self._obtain_screenshot_type(is_element=False,
                                                             inside_a_frame=bool(self._driver.frame_chain),
                                                             stitch_content=False,
                                                             force_fullpage=self.force_full_page_screenshot)
        self._check_window_base(tag, match_timeout, target)

    def check_region(self, region, tag=None, match_timeout=-1, target=None, stitch_content=False):
        # type: (Region, tp.Optional[tp.Text], int, tp.Optional[Target], bool) -> None
        """
        Takes a snapshot of the given region from the browser using the web driver and matches it
        with the expected output. If the current context is a frame, the region is offsetted
        relative to the frame.

        :param region: (Region) The region which will be visually validated. The coordinates are
                         relative to the viewport of the current frame.
        :param tag: (str) Description of the visual validation checkpoint.
        :param match_timeout: (int) Timeout for the visual validation checkpoint (milliseconds).
        :param target: (Target) The target for the check_window call
        :return: None
        """
        logger.info("check_region([%s], '%s')" % (region, tag))
        if region.is_empty():
            raise EyesError("region cannot be empty!")
        self._screenshot_type = self._obtain_screenshot_type(is_element=False,
                                                             inside_a_frame=bool(self._driver.frame_chain),
                                                             stitch_content=stitch_content,
                                                             force_fullpage=self.force_full_page_screenshot,
                                                             is_region=True)
        self._region_to_check = region
        self._check_window_base(tag, match_timeout, target)

    def check_region_by_element(self, element, tag=None, match_timeout=-1, target=None, stitch_content=False):
        # type: (AnyWebElement, tp.Optional[tp.Text], int, tp.Optional[Target], bool) -> None
        """
        Takes a snapshot of the region of the given element from the browser using the web driver
        and matches it with the expected output.

        :param element: (WebElement)  The element which region will be visually validated.
        :param tag: (str) Description of the visual validation checkpoint.
        :param match_timeout: (int) Timeout for the visual validation checkpoint (milliseconds).
        :param target: (Target) The target for the check_window call
        :return: None
        """
        logger.info("check_region_by_element('%s')" % tag)
        self._screenshot_type = self._obtain_screenshot_type(is_element=True,
                                                             inside_a_frame=bool(self._driver.frame_chain),
                                                             stitch_content=stitch_content,
                                                             force_fullpage=self.force_full_page_screenshot)

        self._element_position_provider = ElementPositionProvider(self._driver, element)

        origin_overflow = element.get_overflow()
        element.set_overflow('hidden')

        element_region = self._get_element_region(element)
        self._region_to_check = element_region
        self._check_window_base(tag, match_timeout, target)
        self._element_position_provider = None

        if origin_overflow:
            element.set_overflow(origin_overflow)

    def _get_element_region(self, element):
        #  We use a smaller size than the actual screenshot size in order to eliminate duplication
        #  of bottom scroll bars, as well as footer-like elements with fixed position.
        pl = element.location
        # TODO: add correct values for Safari
        # in the safari browser the returned size has absolute value but not relative as
        # in other browsers
        element_width = element.get_client_width()
        element_height = element.get_client_height()
        border_left_width = element.get_computed_style_int('border-left-width')
        border_top_width = element.get_computed_style_int('border-top-width')
        element_region = Region(pl['x'] + border_left_width,
                                pl['y'] + border_top_width,
                                element_width, element_height)
        return element_region

    def check_region_by_selector(self, by, value, tag=None, match_timeout=-1, target=None, stitch_content=False):
        # type: (tp.Text, tp.Text, tp.Optional[tp.Text], int, tp.Optional[Target], bool) -> None
        """
        Takes a snapshot of the region of the element found by calling find_element(by, value)
        and matches it with the expected output.

        :param by: (By) The way by which an element to be validated should be found (e.g., By.ID).
        :param value: (str) The value identifying the element using the "by" type.
        :param tag: (str) Description of the visual validation checkpoint.
        :param match_timeout: (int) Timeout for the visual validation checkpoint (milliseconds).
        :param target: (Target) The target for the check_window call
        :return: None
        """
        logger.debug("calling 'check_region_by_selector'...")
        # hack: prevent stale element exception by saving viewport value before catching element
        self._driver.get_default_content_viewport_size()
        self.check_region_by_element(self._driver.find_element(by, value), tag,
                                     match_timeout, target, stitch_content)

    def check_region_in_frame_by_selector(self, frame_reference,  # type: FrameReference
                                          by,  # type: tp.Text
                                          value,  # type: tp.Text
                                          tag=None,  # type: tp.Optional[tp.Text]
                                          match_timeout=-1,  # type: int
                                          target=None,  # type: tp.Optional[Target]
                                          stitch_content=False  # type: bool
                                          ):
        # type: (...) -> None
        """
        Checks a region within a frame, and returns to the current frame.

        :param frame_reference: (int/str/WebElement) A reference to the frame in which the region should be checked.
        :param by: (By) The way by which an element to be validated should be found (e.g., By.ID).
        :param value: (str) The value identifying the element using the "by" type.
        :param tag: (str) Description of the visual validation checkpoint.
        :param match_timeout: (int) Timeout for the visual validation checkpoint (milliseconds).
        :param target: (Target) The target for the check_window call
        :return: None
        """
        # TODO: remove this disable
        if self.is_disabled:
            logger.info('check_region_in_frame_by_selector(): ignored (disabled)')
            return
        logger.info("check_region_in_frame_by_selector('%s')" % tag)

        # Switching to the relevant frame
        with self._driver.switch_to.frame_and_back(frame_reference):
            logger.debug("calling 'check_region_by_selector'...")
            self.check_region_by_selector(by, value, tag, match_timeout, target, stitch_content)

    def add_mouse_trigger_by_element(self, action, element):
        # type: (tp.Text, AnyWebElement) -> None
        """
        Adds a mouse trigger.

        :param action: Mouse action (click, double click etc.)
        :param element: The element on which the action was performed.
        """
        if self.is_disabled:
            logger.debug("add_mouse_trigger: Ignoring %s (disabled)" % action)
            return
        # Triggers are activated on the last checked window.
        if self._last_screenshot is None:
            logger.debug("add_mouse_trigger: Ignoring %s (no screenshot)" % action)
            return
        if not self._driver.frame_chain == self._last_screenshot.frame_chain:
            logger.debug("add_mouse_trigger: Ignoring %s (different frame)" % action)
            return
        control = self._last_screenshot.get_intersected_region_by_element(element)
        # Making sure the trigger is within the last screenshot bounds
        if control.is_empty():
            logger.debug("add_mouse_trigger: Ignoring %s (out of bounds)" % action)
            return
        cursor = control.middle_offset
        trigger = MouseTrigger(action, control, cursor)
        self._user_inputs.append(trigger)
        logger.info("add_mouse_trigger: Added %s" % trigger)

    def add_text_trigger_by_element(self, element, text):
        # type: (AnyWebElement, tp.Text) -> None
        """
        Adds a text trigger.

        :param element: The element to which the text was sent.
        :param text: The trigger's text.
        """
        if self.is_disabled:
            logger.debug("add_text_trigger: Ignoring '%s' (disabled)" % text)
            return
        # Triggers are activated on the last checked window.
        if self._last_screenshot is None:
            logger.debug("add_text_trigger: Ignoring '%s' (no screenshot)" % text)
            return
        if not self._driver.frame_chain == self._last_screenshot.frame_chain:
            logger.debug("add_text_trigger: Ignoring %s (different frame)" % text)
            return
        control = self._last_screenshot.get_intersected_region_by_element(element)
        # Making sure the trigger is within the last screenshot bounds
        if control.is_empty():
            logger.debug("add_text_trigger: Ignoring %s (out of bounds)" % text)
            return
        trigger = TextTrigger(control, text)
        self._user_inputs.append(trigger)
        logger.info("add_text_trigger: Added %s" % trigger)



from __future__ import absolute_import

import abc
import re
import typing as tp

from selenium.common.exceptions import WebDriverException

from applitools.core import logger, EyesError, Point
from applitools.utils import ABC

if tp.TYPE_CHECKING:
    from applitools.utils.custom_types import AnyWebDriver, ViewPort, AnyWebElement


class StitchMode(object):
    """
    The type of methods for stitching full-page screenshots.
    """
    Scroll = "Scroll"
    CSS = "CSS"


class PositionProvider(ABC):
    """ Encapsulates page/element positioning """
    _JS_GET_CONTENT_ENTIRE_SIZE = """
        var scrollWidth = document.documentElement.scrollWidth;
        var bodyScrollWidth = document.body.scrollWidth;
        var totalWidth = Math.max(scrollWidth, bodyScrollWidth);
        var clientHeight = document.documentElement.clientHeight;
        var bodyClientHeight = document.body.clientHeight;
        var scrollHeight = document.documentElement.scrollHeight;
        var bodyScrollHeight = document.body.scrollHeight;
        var maxDocElementHeight = Math.max(clientHeight, scrollHeight);
        var maxBodyHeight = Math.max(bodyClientHeight, bodyScrollHeight);
        var totalHeight = Math.max(maxDocElementHeight, maxBodyHeight);
        return [totalWidth, totalHeight];";
    """

    def __init__(self, driver):
        # type: (AnyWebDriver) -> None
        self._driver = driver
        self._states = []  # type: tp.List[Point]

    def _execute_script(self, script):
        # type: (tp.Text) -> tp.List[int]
        return self._driver.execute_script(script)

    @abc.abstractmethod
    def get_current_position(self):
        # type: () -> tp.Optional[Point]
        """
        :return: The current position, or `None` if position is not available.
        """

    @abc.abstractmethod
    def set_position(self, location):
        # type: (Point) -> None
        """
        Go to the specified location.

        :param location: The position to set
        :return:
        """

    def get_entire_size(self):
        # type: () -> ViewPort
        """
        :return: The entire size of the container which the position is relative to.
        """
        try:
            width, height = self._driver.execute_script(self._JS_GET_CONTENT_ENTIRE_SIZE)
        except WebDriverException:
            raise EyesError('Failed to extract entire size!')
        return dict(width=width, height=height)

    def push_state(self):
        """
        Adds the current position to the states list.
        """
        self._states.append(self.get_current_position())

    def pop_state(self):
        """
        Sets the position to be the last position added to the states list.
        """
        self.set_position(self._states.pop())


class ScrollPositionProvider(PositionProvider):
    _JS_GET_CURRENT_SCROLL_POSITION = """
        var doc = document.documentElement;
        var x = window.scrollX || ((window.pageXOffset || doc.scrollLeft) - (doc.clientLeft || 0));
        var y = window.scrollY || ((window.pageYOffset || doc.scrollTop) - (doc.clientTop || 0));
        return [x, y]"""

    def set_position(self, location):
        scroll_command = "window.scrollTo({0}, {1})".format(location.x, location.y)
        logger.debug(scroll_command)
        self._execute_script(scroll_command)

    def get_current_position(self):
        try:
            x, y = self._execute_script(self._JS_GET_CURRENT_SCROLL_POSITION)
            if x is None or y is None:
                raise EyesError("Got None as scroll position! ({},{})".format(x, y))
        except WebDriverException:
            raise EyesError("Failed to extract current scroll position!")
        return Point(x, y)


class CSSTranslatePositionProvider(PositionProvider):
    _JS_TRANSFORM_KEYS = ["transform", "-webkit-transform"]

    def __init__(self, driver):
        super(CSSTranslatePositionProvider, self).__init__(driver)
        self._current_position = Point(0, 0)

    def _set_transform(self, transform_list):
        script = ''
        for key, value in transform_list.items():
            script += "document.documentElement.style['{}'] = '{}';".format(key, value)
        self._execute_script(script)

    def _get_current_transform(self):
        script = 'return {'
        for key in self._JS_TRANSFORM_KEYS:
            script += "'{0}': document.documentElement.style['{0}'],".format(key)
        script += ' }'
        return self._execute_script(script)

    def get_current_position(self):
        return self._current_position.clone()

    def _get_position_from_transform(self, transform):
        data = re.match(r"^translate\(\s*(\-?)([\d, \.]+)px,\s*(\-?)([\d, \.]+)px\s*\)", transform)
        if not data:
            raise EyesError("Can't parse CSS transition")

        x = float(data.group(2))
        y = float(data.group(4))
        minus_x, minus_y = data.group(1), data.group(3)
        if minus_x:
            x *= -1
        if minus_y:
            y *= -1

        return Point(float(x), float(y))

    def set_position(self, location):
        translate_command = "translate(-{}px, -{}px)".format(location.x, location.y)
        logger.debug(translate_command)
        transform_list = dict((key, translate_command) for key in self._JS_TRANSFORM_KEYS)
        self._set_transform(transform_list)
        self._current_position = location.clone()

    def push_state(self):
        """
        Adds the transform to the states list.
        """
        transforms = self._get_current_transform()
        if not all(transforms.values()):
            self._current_position = Point.create_top_left()
        else:
            point = Point(0, 0)
            for transform in transforms.values():
                point += self._get_position_from_transform(transform)
            self._current_position = point
        self._states.append(self._current_position)


class ElementPositionProvider(PositionProvider):

    def __init__(self, driver, element):
        # type: (AnyWebDriver, tp.Optional[AnyWebElement]) -> None
        super(ElementPositionProvider, self).__init__(driver)
        self._element = element

    def get_current_position(self):
        position = Point(self._element.get_scroll_left(), self._element.get_scroll_top())
        logger.info("Current position: {}".format(position))
        return position

    def set_position(self, location):
        logger.info("Scrolling element to {}".format(location))
        self._element.scroll_to(location)
        logger.info("Done scrolling element!")

    def get_entire_size(self):
        try:
            size = {'width': self._element.get_scroll_width(), 'height': self._element.get_scroll_height()}
        except WebDriverException:
            raise EyesError('Failed to extract entire size!')
        logger.info("ElementPositionProvider - Entire size: {}".format(size))
        return size


def build_position_provider_for(stitch_mode,  # type: tp.Text
                                driver,  # type: AnyWebDriver
                                ):
    # type: (...) -> PositionProvider
    if stitch_mode == StitchMode.Scroll:
        return ScrollPositionProvider(driver)
    elif stitch_mode == StitchMode.CSS:
        return CSSTranslatePositionProvider(driver)
    raise ValueError("Invalid stitch mode: {}".format(stitch_mode))



import copy
import typing as tp

from applitools.core import Point, EyesError

if tp.TYPE_CHECKING:
    from applitools.utils.custom_types import FrameReference, RectangleSize

__all__ = ('Frame', 'FrameChain')


class Frame(object):
    """
    Encapsulates a frame/iframe. This is a generic type class,
    and it's actual type is determined by the reference used by the user in
    order to switch into the frame.
    """
    __slots__ = ('reference', 'location', 'outer_size', 'inner_size', 'parent_scroll_position',
                 '_scroll_root_element', '_original_overflow')

    def __init__(self, reference, location, outer_size, inner_size, parent_scroll_position):
        # type: (FrameReference, Point, RectangleSize, RectangleSize, Point) -> None
        """
        :param reference:         The web element for the frame, used as a reference to switch into the frame.
        :param location:          The location of the frame within the current frame.
        :param outer_size:        The frame element outerSize
        (i.e., the outer_size of the frame on the screen, not the internal document outer_size).
        :param inner_size:        The frame element inner outerSize
        (i.e., the outer_size of the frame actual outer_size, without borders).
        :param parent_scroll_position: The scroll location of the frame.
        """
        self._scroll_root_element = None
        self._original_overflow = None

        self.reference = reference
        self.location = location
        self.outer_size = outer_size
        self.inner_size = inner_size
        self.parent_scroll_position = parent_scroll_position

    def __str__(self):
        return "Frame: {}".format(self.reference)


class FrameChain(tp.Sequence[Frame]):
    __slots__ = ('_frames',)

    def __init__(self, frame_chain=None):
        self._frames = []
        if frame_chain is not None:
            assert isinstance(frame_chain, FrameChain), 'Must be a FrameChain'
            self._frames = copy.copy(frame_chain._frames)

    def __iter__(self):
        return iter(self._frames)

    def __getitem__(self, item):
        return self._frames[item]

    def __len__(self):
        return len(self._frames)

    def __str__(self):
        return 'FrameChain with {} frames'.format(len(self))

    def __eq__(self, other):
        cl1, cl2 = len(self._frames), len(other)
        if cl1 != cl2:
            return False
        for i in range(cl1):
            if self._frames[i].id_ != other[i].id_:
                return False
        return True

    def clear(self):
        self._frames = []

    @property
    def size(self):
        return len(self)

    @property
    def peek(self):
        # type: () -> Frame
        return self[-1]

    def push(self, frame):
        # type: (Frame) -> None
        assert isinstance(frame, Frame), 'frame must be instance of Frame!'
        self._frames.append(frame)

    def pop(self):
        # type: () -> Frame
        return self._frames.pop()

    @property
    def current_frame_offset(self):
        # type: () -> Point
        location = Point.create_top_left()
        for frame in self:
            location.offset_by_location(frame.location)
        return location

    @property
    def current_frame_size(self):
        # type: () -> RectangleSize
        return self.peek.inner_size

    @property
    def current_frame_inner_size(self):
        # type: () -> RectangleSize
        return self.peek.outer_size

    @property
    def default_content_scroll_position(self):
        # type: () -> tp.Union[Point, EyesError]
        if len(self) == 0:
            raise EyesError('No frames in frame chain')
        result = self[0].parent_scroll_position
        return Point(result.x, result.y)

    def clone(self):
        # type: () -> FrameChain
        return FrameChain(self)



import codecs
import os
import re
import sys

from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))


def read(filename):
    return codecs.open(path.join(here, filename), 'r', 'utf-8').read()


# preventing ModuleNotFoundError caused by importing lib before installing deps
def get_version():
    with open(os.path.join(os.path.abspath('.'), 'applitools/__version__.py'), 'r') as f:
        try:
            version = re.findall(r"^__version__ = '([^']+)'\r?$",
                                 f.read(), re.M)[0]
        except IndexError:
            raise RuntimeError('Unable to determine version.')
    return version


install_requires = [
    'requests>=2.1.0',
    'selenium>=2.53.0',
    'Pillow>=5.0.0',
    'tinycss2>=0.6.1',
]
# using this way of defining instead of 'typing>=3.5.2; python_version<="3.4"'
# for run on old version of setuptools without issues
if sys.version_info < (3, 5):
    # typing module was added as builtin in Python 3.5
    install_requires.append('typing >= 3.5.2')

setup(
    name='eyes-selenium',
    version=get_version(),
    packages=find_packages(exclude=('tests',)),
    url='http://www.applitools.com',
    license='Apache License, Version 2.0',
    author='Applitools Team',
    author_email='team@applitools.com',
    description='Applitools Eyes SDK For Selenium Python WebDriver',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing"
    ],
    keywords='applitools eyes selenium',
    install_requires=install_requires,
    extras_require={
        'dev':     [
            'bumpversion',
            'flake8',
            'flake8-import-order',
            'flake8-bugbear',
            'mypy',
            'towncrier',
        ],
        'testing': [
            'pytest < 4.0.0',
            'pytest-cov',
            'pytest-xdist',
        ],
    },
    package_data={
        '':           ['README.md', 'samples'],
        'applitools': ['py.typed'],
    },
    project_urls={
        'Bug Reports':                  'https://github.com/applitools/eyes.selenium.python/issues',
        'Selenium Python example':      'https://applitools.com/resources/tutorial/selenium/python#step-2',
        'Python Appium native example': 'https://applitools.com/resources/tutorial/appium/native_python#step-2',
        'Python Appium web example':    'https://applitools.com/resources/tutorial/appium/python#step-2',
        'Source':                       'https://github.com/applitools/eyes.selenium.python',
    },
)



from __future__ import absolute_import

import base64
import typing as tp

from selenium.common.exceptions import WebDriverException

from applitools.core import EyesScreenshot, EyesError, Point, Region, OutOfBoundsError
from applitools.utils import image_utils
from applitools.selenium import eyes_selenium_utils
from applitools.selenium.frames import FrameChain

if tp.TYPE_CHECKING:
    from PIL import Image
    from applitools.utils.custom_types import ViewPort
    from applitools.selenium import EyesWebDriver


class EyesWebDriverScreenshot(EyesScreenshot):

    @staticmethod
    def create_from_base64(screenshot64, driver):
        """
        Creates an instance from the base64 data.

        :param screenshot64: The base64 representation of the png bytes.
        :param driver: The webdriver for the session.
        """
        return EyesWebDriverScreenshot(driver, screenshot64=screenshot64)

    @staticmethod
    def create_from_image(screenshot, driver):
        # type: (Image.Image, EyesWebDriver) -> EyesWebDriverScreenshot
        """
        Creates an instance from the base64 data.

        :param screenshot: The screenshot image.
        :param driver: The webdriver for the session.
        """
        return EyesWebDriverScreenshot(driver, screenshot=screenshot)

    def __init__(self, driver, screenshot=None, screenshot64=None,
                 is_viewport_screenshot=None, frame_location_in_screenshot=None):
        # type: (EyesWebDriver, Image.Image, None, tp.Optional[bool], tp.Optional[Point]) -> None
        """
        Initializes a Screenshot instance. Either screenshot or screenshot64 must NOT be None.
        Should not be used directly. Use create_from_image/create_from_base64 instead.

        :param driver: EyesWebDriver instance which handles the session from which the screenshot
                    was retrieved.
        :param screenshot: image instance. If screenshot64 is None,
                                    this variable must NOT be none.
        :param screenshot64: The base64 representation of a png image. If screenshot
                                     is None, this variable must NOT be none.
        :param is_viewport_screenshot: Whether the screenshot object represents a
                                                viewport screenshot or a full screenshot.
        :param frame_location_in_screenshot: The location of the frame relative
                                                    to the top,left of the screenshot.
        :raise EyesError: If the screenshots are None.
        """
        if screenshot is None and screenshot64 is None:
            raise EyesError("both screenshot and screenshot64 are None!")

        if screenshot64:
            screenshot = image_utils.image_from_bytes(base64.b64decode(screenshot64))

        # initializing of screenshot
        super(EyesWebDriverScreenshot, self).__init__(image=screenshot)

        self._driver = driver
        self._viewport_size = driver.get_default_content_viewport_size(force_query=False)  # type: ViewPort

        self._frame_chain = driver.frame_chain.clone()
        if self._frame_chain:
            chain_len = len(self._frame_chain)
            self._frame_size = self._frame_chain[chain_len - 1].outer_size
        else:
            try:
                self._frame_size = driver.get_entire_page_size()
            except WebDriverException:
                # For Appium, we can't get the "entire page size", so we use the viewport size.
                self._frame_size = self._viewport_size
        # For native Appium Apps we can't get the scroll position, so we use (0,0)
        try:
            self._scroll_position = driver.get_current_position()
        except (WebDriverException, EyesError):
            self._scroll_position = Point(0, 0)
        if is_viewport_screenshot is None:
            is_viewport_screenshot = (self._screenshot.width <= self._viewport_size['width']
                                      and self._screenshot.height <= self._viewport_size['height'])
        self._is_viewport_screenshot = is_viewport_screenshot
        if frame_location_in_screenshot is None:
            if self._frame_chain:
                frame_location_in_screenshot = EyesWebDriverScreenshot \
                    .calc_frame_location_in_screenshot(self._frame_chain, is_viewport_screenshot)
            else:
                # The frame is the default content
                frame_location_in_screenshot = Point(0, 0)
                if self._is_viewport_screenshot:
                    frame_location_in_screenshot.offset(-self._scroll_position.x,
                                                        -self._scroll_position.y)
        self._frame_location_in_screenshot = frame_location_in_screenshot
        self._frame_screenshot_intersect = Region(frame_location_in_screenshot.x,
                                                  frame_location_in_screenshot.y,
                                                  self._frame_size['width'],
                                                  self._frame_size['height'])
        self._frame_screenshot_intersect.intersect(Region(width=self._screenshot.width,
                                                          height=self._screenshot.height))

    @staticmethod
    def calc_frame_location_in_screenshot(frame_chain, is_viewport_screenshot):
        first_frame = frame_chain[0]
        location_in_screenshot = Point(first_frame.location['x'], first_frame.location['y'])
        # We only need to consider the scroll of the default content if the screenshot is a
        # viewport screenshot. If this is a full page screenshot, the frame location will not
        # change anyway.
        if is_viewport_screenshot:
            location_in_screenshot.x -= first_frame.parent_scroll_position.x
            location_in_screenshot.y -= first_frame.parent_scroll_position.y
        # For inner frames we must calculate the scroll
        inner_frames = frame_chain[1:]
        for frame in inner_frames:
            location_in_screenshot.x += frame.location['x'] - frame.parent_scroll_position.x
            location_in_screenshot.y += frame.location['y'] - frame.parent_scroll_position.y
        return location_in_screenshot

    @property
    def frame_chain(self):
        return self._frame_chain

    def get_base64(self):
        if not self._screenshot64:
            self._screenshot64 = image_utils.get_base64(self._screenshot)
        return self._screenshot64

    def get_location_relative_to_frame_viewport(self, location):
        result = {'x': location['x'], 'y': location['y']}
        if self._frame_chain or self._is_viewport_screenshot:
            result['x'] -= self._scroll_position.x
            result['y'] -= self._scroll_position.y
        return result

    def get_sub_screenshot_by_region(self, region):
        sub_screenshot_region = self.get_intersected_region(region)
        if sub_screenshot_region.is_empty():
            raise OutOfBoundsError("Region {0} is out of bounds!".format(region))
        # If we take a screenshot of a region inside a frame, then the frame's (0,0) is in the
        # negative offset of the region..
        sub_screenshot_frame_location = Point(-region.left, -region.top)
        # FIXME Calculate relative region location? (same as the java version)
        screenshot = image_utils.get_image_part(self._screenshot, sub_screenshot_region)
        return EyesWebDriverScreenshot(self._driver, screenshot,
                                       is_viewport_screenshot=self._is_viewport_screenshot,
                                       frame_location_in_screenshot=sub_screenshot_frame_location)

    def get_element_region_in_frame_viewport(self, element):
        location, size = element.location, element.size

        relative_location = self.get_location_relative_to_frame_viewport(location)

        x, y = relative_location['x'], relative_location['y']
        width, height = size['width'], size['height']
        # We only care about the part of the element which is in the viewport.
        if x < 0:
            diff = -x
            # IMPORTANT the diff is between the original location and the viewport's bounds.
            width -= diff
            x = 0
        if y < 0:
            diff = -y
            height -= diff
            y = 0

        if width <= 0 or height <= 0:
            raise OutOfBoundsError("Element's region is outside the viewport! [(%d, %d) %d x %d]" %
                                   (location['x'], location['y'], size['width'], size['height']))

        return Region(x, y, width, height)

    def get_intersected_region(self, region):
        region_in_screenshot = region.clone()
        region_in_screenshot.left += self._frame_location_in_screenshot.x
        region_in_screenshot.top += self._frame_location_in_screenshot.y
        region_in_screenshot.intersect(self._frame_screenshot_intersect)
        return region_in_screenshot

    def get_viewport_screenshot(self):
        # if screenshot if full page
        if not self._is_viewport_screenshot and not eyes_selenium_utils.is_mobile_device(self._driver):
            return self.get_sub_screenshot_by_region(
                Region(top=self._scroll_position.y, height=self._viewport_size['height'],
                       width=self._viewport_size['width']))
        return self



from __future__ import absolute_import

import abc
import typing as tp

from PIL import Image

from applitools.utils import ABC, image_utils, argument_guard

if tp.TYPE_CHECKING:
    from applitools.utils.custom_types import AnyWebElement, Num
    from applitools.selenium.frames import FrameChain
    from applitools.selenium.webelement import EyesWebElement
    from .geometry import Point, Region

    T = tp.TypeVar('T', bound='EyesScreenshot')

__all__ = ('EyesScreenshot',)


class EyesScreenshot(ABC):
    """
    Base class for handling screenshots.
    """

    def __init__(self, image):
        # type: (Image.Image) -> None
        argument_guard.is_a(image, Image.Image)
        self._screenshot = image

    @staticmethod
    @abc.abstractmethod
    def calc_frame_location_in_screenshot(frame_chain, is_viewport_screenshot):
        # type: (FrameChain, tp.Optional[bool]) -> Point
        """
        :param frame_chain: List of the frames.
        :param is_viewport_screenshot: Whether the viewport is a screenshot or not.
        :return: The frame location as it would be on the screenshot. Notice that this value
            might actually be OUTSIDE the screenshot (e.g, if this is a viewport screenshot and
            the frame is located outside the viewport). This is not an error. The value can also
            be negative.
        """

    @abc.abstractmethod
    def get_base64(self):
        # type: () -> tp.Text
        """
        Returns a base64 screenshot.

        :return: The base64 representation of the png.
        """

    @abc.abstractmethod
    def get_intersected_region(self, region):
        # type: (Region) -> Region
        """
        Gets the intersection of the region with the screenshot image.

        :param region: The region in the frame.
        :return: The part of the region which intersects with
            the screenshot image.
        """

    @abc.abstractmethod
    def get_location_relative_to_frame_viewport(self, location):
        # type: (tp.Dict[tp.Text, Num]) -> tp.Dict[tp.Text, Num]
        """
        Gets the relative location from a given location to the viewport.

        :param location: A dict with 'x' and 'y' keys representing the location we want
            to adjust.
        :return: A location (keys are 'x' and 'y') adjusted to the current frame/viewport.
        """

    @abc.abstractmethod
    def get_element_region_in_frame_viewport(self, element):
        # type: (AnyWebElement) -> Region
        """
        Gets The element region in the frame.

        :param element: The element to get the region in the frame.
        :return: The element's region in the frame with scroll considered if necessary
        """

    @abc.abstractmethod
    def get_viewport_screenshot(self):
        # type: () -> T
        """
        Always return viewport size screenshot
        """

    @abc.abstractmethod
    def get_sub_screenshot_by_region(self, region):
        # type: (Region) -> T
        """
        Gets the region part of the screenshot image.

        :param region: The region in the frame.
        :return: A screenshot object representing the given region part of the image.
        """

    @property
    @abc.abstractmethod
    def frame_chain(self):
        # type: () -> FrameChain
        """
        Returns a copy of the fram chain.

        :return: A copy of the frame chain, as received by the driver when the screenshot was
            created.
        """

    def get_bytes(self):
        # type: () -> bytes
        """
        Returns the bytes of the screenshot.

        :return: The bytes representation of the png.
        """
        return image_utils.get_bytes(self._screenshot)

    def get_intersected_region_by_element(self, element):
        # type: (EyesWebElement) -> Region
        """
        Gets the intersection of the element's region with the screenshot image.

        :param element: The element in the frame.
        :return: The part of the element's region which intersects with
            the screenshot image.
        """
        element_region = self.get_element_region_in_frame_viewport(element)
        return self.get_intersected_region(element_region)

    def get_sub_screenshot_by_element(self, element):
        # type: (EyesWebElement) -> T
        """
        Gets the element's region part of the screenshot image.

        :param element: The element in the frame.
        :return: A screenshot object representing the element's region part of the
            image.
        """
        element_region = self.get_element_region_in_frame_viewport(element)
        return self.get_sub_screenshot_by_region(element_region)
